/* Copyright (C) 1989 Aladdin Enterprises.  All rights reserved.
   Distributed by Free Software Foundation, Inc.

This file is part of Ghostscript.

Ghostscript is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
to anyone for the consequences of using it or for whether it serves any
particular purpose or works at all, unless he says so in writing.  Refer
to the Ghostscript General Public License for full details.

Everyone is granted permission to copy, modify and redistribute
Ghostscript, but only under the conditions described in the Ghostscript
General Public License.  A copy of this license is supposed to have been
given to you along with Ghostscript so you can know your rights and
responsibilities.  It should be in a file named COPYING.  Among other
things, the copyright notice and this notice must be preserved on all
copies.  */

/* ialloc.c */
/* Memory allocator for GhostScript interpreter */
#include <stdio.h>		/* for NULL */
#include "std.h"
#include "malloc_.h"
#include "memory_.h"
#include "alloc.h"

extern char Debug[128];

/* Round up all sizes so objects are aligned. */
#define alloc_mod 4
#define alloc_mask (alloc_mod-1)
#define alloc_round(siz) (uint)(((siz) + alloc_mask) & -alloc_mod)

/* Note that allocation takes place both from the bottom up */
/* (dynamic objects) and from the top down. */
static byte *alloc_base;
static byte *alloc_bot;			/* bottom of free area */
static byte *alloc_top;			/* top of free area */
static byte *alloc_limit;
static uint alloc_chunk_size, alloc_big_size;
/* Cumulative statistics */
static long a_used, a_total;

/* Chain together any malloc'ed objects */
static char *malloc_chain;

/* Debugging printout */
#ifdef DEBUG
#  define alloc_print(tag, blk, sz)\
	if ( Debug['a'] ) printf("[a%c]%lx(%u) ", tag, (ulong)blk, sz)
#else
#  define alloc_print(tag, blk, sz)
#endif

/* ------ Initialize/status ------ */

/* Initialize the allocator */
void
alloc_init(uint chunk_size)
{	alloc_base = alloc_bot = alloc_top = alloc_limit = NULL;
	alloc_chunk_size = chunk_size;
	alloc_big_size = chunk_size / 3;
	a_used = a_total = 0;
	malloc_chain = 0;
}

/* Return the status of the allocator: space used, total space. */
void
alloc_status(long *pused, long *ptotal)
{	*pused = (alloc_bot - alloc_base) + (alloc_limit - alloc_top) + a_used;
	*ptotal = alloc_limit - alloc_base + a_total;
}

/* ------ Ordinary allocation ------ */

/* Allocate an object.  Return 0 if not enough room. */
char *
alloc(uint size)
{	uint left = alloc_top - alloc_bot;
	uint block_size;
	if ( size >= alloc_big_size )
	   {	/* Large object, do a separate malloc. */
		char *mblock = malloc(size + sizeof(char *));
		if ( mblock != NULL )
		   {	char *block = mblock + sizeof(char *);
		   	alloc_print('*', block, size);
			*(char **)mblock = malloc_chain;
			malloc_chain = mblock;
			return block;
		   }
	   }
	block_size = alloc_round(size);
	if ( block_size > left )
	   {	if ( !alloc_add_chunk() )
		   {	alloc_print('~', (ulong)0, size);
#ifdef DEBUG
			gs_dump_C_stack();
#endif
			return 0;
		   }
	   }
	alloc_top -= block_size;
	alloc_print(' ', alloc_top, size);
	return (char *)alloc_top;
}

/* Free an object, if possible. */
void
alloc_free(char *cobj, uint size)
{	uint block_size = alloc_round(size);
#define obj ((byte *)cobj)
	if ( obj == alloc_top )
		alloc_top += block_size;
	else if ( obj + block_size == alloc_bot )
		alloc_bot = (byte *)obj;
	else if ( (ptr_ord_t)obj < (ptr_ord_t)alloc_base ||
		  (ptr_ord_t)obj >= (ptr_ord_t)alloc_limit )
	   {	/* Might be in another segment, or allocated with */
		/* malloc.  Check the malloc chain. */
		char **prev;
		char *mblock;
		for ( prev = &malloc_chain; *prev != 0; prev = (char **)mblock )
		   {	mblock = *prev;
			if ( mblock + sizeof(char *) == cobj )
			   {	*prev = *(char **)mblock;
				free(mblock);
				alloc_print('^', obj, size);
				return;
			   }
		   }
		alloc_print('?', obj, size);
		return;
	   }
	else
	   {	alloc_print('=', obj, size);
#ifdef DEBUG
		if ( Debug['a'] ) gs_dump_C_stack();
#endif
		return;
	   }
	alloc_print('-', obj, size);
#undef obj
}

/* ------ Dynamic allocation ------ */

/* Allocate a dynamic object.  Return 0 if not enough room. */
byte *
alloc_dynamic(uint size)
{	/* We take the easy way out here.... */
	byte *obj = (byte *)alloc(size);
	uint block_size = alloc_round(size);
	if ( obj != alloc_top ) return obj;	/* 0, or separate malloc */
	/* "Move" the object from the top to the bottom. */
	alloc_top += block_size;
	obj = alloc_bot;
	alloc_bot += block_size;
	alloc_print('<', obj, size);
	return obj;
}

/* Grow a dynamic object.  This may require moving it to a new chunk. */
/* Return 0 if not enough room. */
byte *
alloc_grow(byte *obj, uint old_size, uint new_size)
{	byte *nobj;
	uint old_block_size = alloc_round(old_size);
	uint new_block_size = alloc_round(new_size);
	if ( obj + old_block_size == alloc_bot )
	   {	/* Might be able to grow in place */
		if ( new_block_size <= alloc_top - obj )
		   {	alloc_bot = obj + new_block_size;
			alloc_print('+', obj, new_size);
			return obj;
		   }
	   }
	/* Can't grow in place.  Allocate a new object and copy. */
	nobj = alloc_dynamic(new_size);
	if ( nobj == 0 ) return 0;
	memcpy(nobj, obj, old_size);
	alloc_free((char *)obj, old_size);
	alloc_print('&', obj, new_size);
	return nobj;
}

/* Shrink a dynamic object. */
byte *
alloc_shrink(byte *obj, uint old_size, uint new_size)
{	uint old_block_size = alloc_round(old_size);
	uint new_block_size = alloc_round(new_size);
	if ( obj + old_block_size == alloc_bot )
		alloc_bot = obj + new_block_size;
	alloc_print('.', obj, new_size);
	return obj;
}

/* Shrink a dynamic object and make it permanent. */
byte *
alloc_shrink_copy(byte *obj, uint old_size, uint new_size)
{	uint old_block_size = alloc_round(old_size);
	byte *perm;
	if ( obj + old_block_size == alloc_bot )
	   {	/* Object is at end of current segment, */
		/* so there is definitely room. */
		uint new_block_size = alloc_round(new_size);
		perm = alloc_top -= new_block_size;
		memcpy(perm, obj, new_size);
		alloc_bot = obj;
	   }
	else
	   {	/* Object is somewhere else. */
		/* Just leave it where it is. */
		perm = obj;
	   }
	alloc_print('>', obj, new_size);
	return perm;
}

/* ------ Private routines ------ */

/* Allocate a new chunk.  Return true if successful. */
static int
alloc_add_chunk()
{	char *chunk = malloc(alloc_chunk_size);
#ifdef DEBUG
if ( Debug['a'] )
	printf("[a]%lx@%u ", (ulong)chunk, alloc_chunk_size);
#endif
	if ( chunk == NULL ) return 0;
	/* Update statistics for the old chunk */
	alloc_status(&a_used, &a_total);
	alloc_base = alloc_bot = (byte *)chunk;
	alloc_limit = alloc_top = (byte *)(chunk + alloc_chunk_size);
	return 1;
}
